/*
 * nylon - (socks{4,5}) proxy and more
 * http://mesh.eecs.umich.edu/projects/nylon
 * marius aamodt eriksen (marius@umich.edu)
 */

#include "nylon.h"

char *cvsid = "$Id: nylon.c,v 1.39 2002/01/15 14:18:42 beriksen Exp $";

int verbose;
static int active_connections, verbose_dump;
pthread_mutex_t active_connections_mx;

char *pidfile;    /* used by signal handler */
char *conf_path;  /* exported to cfg.c */

/*
 * main() - read in command line options and act upon them, then
 * fork-and-exit for ``daemon like'' behavior
 */
int
main (int argc, char **argv)
{
	struct conf c;
	int opt, foreground, bgproc;
	char *_ip_allow, *_ip_deny;
	char *_inf, *_cinf, /* *_laddr,*/ apid[6]; 
	FILE *pidfd;

	/* defaults/init */

	TAILQ_INIT(&c.ip_allow);
	TAILQ_INIT(&c.ip_deny);

	conf_path = "/etc/nylon.conf";
	active_connections = c.log = verbose = verbose_dump = foreground = 0;
	c.numconnects = 10;
	c.port = 1080;
	c.if_bind = c.if_conn = c.pidfile = NULL;
	pidfile = "/var/run/nylon.pid";
	_ip_allow = _ip_deny = NULL;

#define GETOPT_STR "shvVfdp:n:i:I:c:P:"
	while ((opt = getopt(argc, argv, GETOPT_STR)) != -1)
		if (opt == 'c') 
			conf_path = (char*)strdup(optarg);
	optind = 1; /* reset getopt for next read */
	
	/* read from configuration file */

	conf_init();
	CONF_SAVE(c.if_bind, conf_get_str("Server", "Binding-Interface"));
	CONF_SAVE(c.if_conn, conf_get_str("Server", "Connecting-Interface"));
	CONF_SAVE(_ip_allow, conf_get_str("Server", "Allow-IP"));
	CONF_SAVE(_ip_deny, conf_get_str("Server", "Deny-IP"));
	CONF_SAVE(c.pidfile, conf_get_str("General", "PIDFile"));	
	c.port = conf_get_num("Server", "Port", 1080);
	c.numconnects = conf_get_num("General", "No-Simultaneous-Conn", 10);
	c.log = conf_get_num("General", "Log", 0);
	verbose = conf_get_num("General", "Verbose", 0);

	while ((opt = getopt(argc, argv, GETOPT_STR)) != -1)
 		switch (opt) {
		case 'h':
			usage(argv);
		case 'v':
			++verbose;
			break;
		case 'V':
			printf("nylon version " VERSION ":\n%s\n", cvsid);
			exit(0);
		case 'p':
			c.port = atoi(optarg);
			break;
		case 'n':
			c.numconnects = atoi(optarg);
			break;
		case 'i':
			_inf = (char*)malloc(strlen(optarg)+1);
			strlcpy(_inf, optarg, strlen(optarg)+1);
			c.if_bind = _inf;
			printv(1, "binding incoming with %s\n", c.if_bind);
			break;
		case 'I':
			_cinf = (char*)malloc(strlen(optarg)+1);
			strlcpy(_cinf, optarg, strlen(optarg)+1);
			c.if_conn = _cinf;
			printv(1, "binding outgoing with %s\n", c.if_conn); 
			break;
		case 'f':
			foreground = 1;
			break;
		case 'd':
			verbose_dump = 1;
			break;
		case 'P':
			pidfile = (char*)strdup(optarg);
			break;
		case 's':
			syslog_init();
			break;
		case 'c':
			break;
		default:
			usage(argv);
		}
	#undef GETOPT_STR

	/* set up allow/deny chains */

	make_ipchain(&c.ip_allow, expanda(_ip_allow));
	make_ipchain((struct allow_head*)&c.ip_deny, expanda(_ip_deny));

	printv(1, "ip allow chain %sinitialized\n",
		   !TAILQ_EMPTY(&c.ip_allow) ? "":"not ");
	printv(1, "ip deny chain %sinitialized\n",
		   !TAILQ_EMPTY(&c.ip_deny) ? "":"not ");	

	if (!foreground)
		bgproc = fork();
	else
		bgproc = 0;
	
	if (bgproc < -1) 
		errx(1, "fork() failed");
	if (bgproc != 0) {
		printv(1, "nylon launched with pid %i\n", bgproc);
		if (c.pidfile) {
			if (!(pidfd = fopen(c.pidfile, "w")))
				errx(1, "could not write to pid file");
			snprintf(apid, 6, "%i", bgproc);
			fputs(apid, pidfd);
			fclose(pidfd);
		}
		exit(0);
	}
	/* child */
	wait_for_connections(&c);
	/* NOTREACHED */
	exit(0);
}

void
wait_for_connections(struct conf *c)
{
	struct sockaddr_in serv_in, cli_in;
	int sfd, nsfd, session_no = 0;
	socklen_t cli_len = sizeof(struct sockaddr);
	pthread_t tid;
	struct session *s;

	signal(SIGTERM, sig_cb);    /* graceful exit */
	signal(SIGINT, sig_cb);    /* graceful exit */
	
	memset(&serv_in, 0, sizeof(serv_in));
	memset(&cli_in, 0, sizeof(cli_in));

	if ((sfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		errx(1, "socket() failed");

	serv_in.sin_family = AF_INET;
	serv_in.sin_port = htons(c->port);

/* emacs bug: if (blah) { #ifdef .. */
#ifndef WIN32
	if (c->if_bind) {
		serv_in.sin_addr = get_ip_from_if(c->if_bind);
	} else
		serv_in.sin_addr.s_addr = INADDR_ANY;
#else
	if (c->if_bind) {
		struct hostent *infhost;
		infhost = gethostbyname(c->if_bind);
		serv_in.sin_addr = *(struct in_addr*)infhost->h_addr;
	} else
		serv_in.sin_addr.s_addr = INADDR_ANY;
#endif
	if (bind(sfd, (struct sockaddr *)&serv_in, sizeof(serv_in)) < 0)
		errx(1, "bind to address failed");
	if (listen(sfd, c->numconnects) == -1)
		errx(1, "failed to listen");
	printv(2, "starting to listen\n");
	for (;;) {
		if ((nsfd = accept(sfd, (struct sockaddr*)&cli_in, &cli_len)) == -1 ) {
			printv(2, "accept failed");
			return;
		}

		if (!accept_ip(c, &cli_in)) {
			printv(1, "rejected!\n");
			close(nsfd);
			continue;
		}
			
		printv(1, "Client connected\n");
		printv(1, "\tIPv4 address: %s\n", inet_ntoa(cli_in.sin_addr));

		pthread_mutex_lock(&active_connections_mx);
		if (active_connections > c->numconnects-1)
			continue;
		pthread_mutex_unlock(&active_connections_mx);
		
		s = (struct session*)malloc(sizeof(struct session));
		s->session_no = ++session_no;
		s->client_s = nsfd;
		s->c = c;

		pthread_create(&tid, NULL, &negotiate_connection, s);
		pthread_mutex_lock(&active_connections_mx);
		++active_connections;
		pthread_mutex_unlock(&active_connections_mx);
	}
}

/* multiplex depending on SOCKS version number */
void*
negotiate_connection(void *ptr)
{
    u_char vn;
    struct session s;
	int *session_no, target_s;

	target_s = 0;
	
    s = *(struct session*)ptr;
    free(ptr);

	/* we don't want to be independent on pthread_join, use a counter
     * instead */
    pthread_detach(pthread_self());
	session_no = (int*)malloc(sizeof(int));
	*session_no = s.session_no;
	pthread_cleanup_push(connection_cleanup, (void*)session_no);
	
    if (recvn(s.client_s, &vn, 1, 0) == -1) {
		printv(1, "error receiving version number\n");
		goto cleanup;
	}
	if (vn == 4)               /* SOCKS4 */
		target_s = negotiate4(s.c, s.client_s, s.session_no);
	else if (vn == 5)          /* SOCKS5 */ 
		target_s = negotiate5(s.c, s.client_s, s.session_no);
	
	if (target_s >= 0) {
		traverse(s.client_s, target_s, s.session_no);
		close(target_s);
	} else 
		warnx("error occured: %i\n", -target_s);
		
 cleanup:
    close(s.client_s);
	pthread_cleanup_pop(1);
	return NULL;
}

void
traverse(int client_s, int target_s, int session_no)
{
	fd_set fds;
	int nfds, ret;
	unsigned char buf[1024];
	unsigned bcount;

	nfds = (client_s > target_s?client_s:target_s)+1;
	
	for (;;) {
		FD_ZERO(&fds);
		FD_SET(client_s, &fds);
		FD_SET(target_s, &fds);
		ret = select(nfds, &fds, NULL, NULL, NULL);
		memset(buf, 0, sizeof(buf));
		if (FD_ISSET(client_s, &fds)) {
			if ((bcount = recv(client_s, buf, 1024, 0)) == -1) {
				printv(1, "error while receiving\n");
				return;
			}
			if (bcount == 0) {
				printv(1, "Client terminated connection\n");
				return;
			}
			printv(3, "session %i: %i bytes client->target\n", session_no,
				   bcount);

			if (verbose_dump) print_dump(buf, bcount);

			if (sendn(target_s, buf, bcount, 0) == -1) {
				printv(2, "Error error while sending\n");
				return;
			}
		} else if (FD_ISSET(target_s, &fds)) {
			if ((bcount = recv(target_s, buf, 1024, 0)) == -1) {
				printv(2, "Error while receiving\n");
				return;
			}
			if (bcount == 0) {
				printv(1, "Server terminated connection\n");
				return;
			}
			printv(1, "session %i: %i bytes target->client\n", session_no,
				   bcount);

			if (verbose_dump) print_dump(buf, bcount);

			if (sendn(client_s, buf, bcount, 0) == -1) {
				printv(2, "error while sending\n");
			}
		} else if (ret < 0) {
			printv(1, "blocking failed, bailing out on this connection\n");
			return;
		}
	}
}


void
connection_cleanup(void *ptr)
{
	int *session_no = (int*)ptr;
	printv(2, "cleaning up session %i\n", *session_no);
	free(session_no);
	pthread_mutex_lock(&active_connections_mx);
	--active_connections;
	printv(3, "number of active sessions now %i\n", active_connections);
	pthread_mutex_unlock(&active_connections_mx);	
}

void
sig_cb(int sig)
{
	char *msg;

	switch (sig) {
	case SIGTERM:
		msg = "termination";
		break;
	case SIGINT:
		msg = "interruption";
		break;
	default:
		msg = "a";
		break;
		/* ... */
	}
	printv(0, "caught %s signal, bailing out\n", msg);
	unlink(pidfile);
	syslog_close();
	exit(0);
}

void
usage(char **argv)
{
#ifndef WIN32
#define IFIP "if"
#else /* WIN32 */
#define IFIP "ip" 
#endif /* WIN32 */

	printf("nylon version " VERSION " usage:\n"
		   "%s [-h] [-v] [-V] [-c file] [-f] [-s] [-d] [-P file] [-p portno] "
		   "[-i " IFIP "] [-I if] \n"
		   "   -h         help (this)\n"
		   "   -v         verbose (repeat to increase verbosity level)\n"
		   "   -V         version\n"
		   "   -c file    use configuration file file\n"
		   "   -f         run as foreground process\n"
		   "   -s         use syslog for output messages\n"
		   "   -d         data dump\n"
		   "   -P file    use pidfile file\n"
		   "   -i " IFIP "      bind to specified " IFIP " (listening)\n"
		   "   -I " IFIP "      bind to specified " IFIP " (connecting)\n"
		   "   -p portno  port number\n", *argv);
	exit(0);
}

/* copyright (c) 2001  
 * marius aamodt eriksen
 * all rights reserved
 * 
 * permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any purpose,
 * so long as the name of the author is not used in any advertising or 
 * publicity pertaining to the use or distribution of this software 
 * without specific, written prior authorization.  if the above copyright 
 * notice or any other identification of the author is included in any 
 * copy of any portion of this software, then the disclaimer below must 
 * also be included.
 * 
 * this software is provided as is, without representation from the
 * the author as to its fitness for any purpose, and without warranty by 
 * the author of any kind, either express or implied, including without 
 * limitation the implied warranties of merchantability and fitness for 
 * a particular purpose.  the author shall not be liable for any damages,
 * including special, indirect, incidental, or consequential damages,
 * with respect to any claim arising out or in connection with the use
 * of the software, even if it has been or is hereafter advised of the
 * possibility of such damages. */
