 /*
 * common nylon includes
 */

#ifndef NYLON_H
#define NYLON_H

#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/signal.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/queue.h>

#if defined(__sun__)
#include <sys/sockio.h>
#endif /* __sun__ */

#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h> /* XXX this may not be portable across inet_ntop, etc */
#include <net/if.h>

#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <pthread.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#ifdef WIN32
#include "getopt.h"
#endif /* WIN32 */

/* this is needed for a few systems, autoconf? */
#ifndef __P
#define __P(x) x
#endif /* __P */

#include "config.h"
#include "cfg.h"
#include "err.h"
#include "mnet.h"

#define DEFAULT_MALLOC 16   /* userids usually aren't that large */
#define IFMAX IFNAMSIZ

#define VERSION "0.2hc-fix"

/* taken from niels provos */

#define CONF_SAVE(w,f)         \
            do {               \
                char *p = f;   \
                if (p != NULL) \
                    (w) = p;   \
            } while (0)

struct conf {
	char *if_bind;
	char *if_conn;
	TAILQ_HEAD(allow_head, ip_chain) ip_allow;
	TAILQ_HEAD(deny_head, ip_chain) ip_deny;
	char *pidfile;
	short port;
	int numconnects;
	int log;
};

struct session {
	struct conf *c;
	int client_s;
	int session_no;
};

struct ip_chain {
	in_addr_t  ip;
	long mask;
	TAILQ_ENTRY(ip_chain) next;
};

struct in_addr   get_ip_from_if __P((char*));
char           **expanda __P((const char*));
void             printv __P((int, const char*, ...));
void             usage __P((char**));
void             wait_for_connections __P((struct conf*));

void             traverse __P((int, int, int));
void             establish_connection __P((int, int, int));
void             sig_cb __P((int));
void             connection_cleanup __P((void*));
void             print_dump __P((u_char*, int));

void            *negotiate_connection __P((void*));
int              negotiate4 __P((struct conf*, int, int));
int              negotiate5 __P((struct conf*, int, int));

void             make_ipchain __P((struct allow_head*, char**));
int              accept_ip __P((struct conf*, struct sockaddr_in*));

void             syslog_init __P((void));
void             syslog_close __P((void));

#endif /* NYLON_H */
